class Column:
    def __init__(self, name, table, column_data):
        self.name = name
        self.table = table
        self.data = column_data

    @property
    def object_name(self):
        return f"{self.table}.{self.name}"





import json
from enum import Enum, auto
from typing import List, Dict


class ParseListStrategy(Enum):
    TO_STR = auto()
    TO_ROWS = auto()
    TO_COLUMNS = auto()


class ParseDictStrategy(Enum):
    TO_STR = auto()
    TO_COLUMNS = auto()


def convert_str_to_parse_list_option(input_str: str) -> ParseListStrategy:
    if input_str.lower() == "to_str":
        return ParseListStrategy.TO_STR
    elif input_str.lower() == "to_rows":
        return ParseListStrategy.TO_ROWS
    elif input_str.lower() == "to_columns":
        return ParseListStrategy.TO_COLUMNS


def convert_str_to_parse_dict_option(input_str: str) -> ParseDictStrategy:
    if input_str.lower() == "to_str":
        return ParseDictStrategy.TO_STR
    elif input_str.lower() == "to_columns":
        return ParseDictStrategy.TO_COLUMNS


DEFAULT_PARSE_LIST = ParseListStrategy.TO_ROWS
DEFAULT_PARSE_DICT = ParseDictStrategy.TO_COLUMNS


class ParserConfigurationError(Exception):
    pass


class ParserConfiguration:
    """
            Args:
                json_configuration_file:
                    Optional[str] : Path to JSON configuration file for the JSONParser settings
                configuration_dict:
                    Optional[Dict] : Configuration of the JSONParser settings in Dict format
                parent_table:
                    Dict : Parent table definition containing the key value pair of parent object name and the
                    resulting csv name. e.g. {"parent_object": "parent_table_name.csv"}
                child_tables:
                    Optional[dict]: Child table definitions containing the key value pairs of child object and their
                    corresponding csv name. Keys being the object name (with all parent objects separated by periods),
                    and the value being the name of the output table. e.g. {"parent_object.child_object": "child.csv"}
                table_primary_keys:
                    Optional[dict]: definition of all primary keys of all tables. Key value pairs where
                    Keys being the object name (with all parent objects separated by periods) and values being
                    key value pairs of objects that should be primary keys of the table and values being the resulting
                    names of the columns in the CSV
                    e.g. {"order.order-item" : {"order.id" : "order_id", "order.order-items.item_id" : "item_id"}}
                table_column_mapping:
                    Optional[Dict[str, Dict[str, str]]] : definition of objects is tables and their mapping to columns.
                    The input is a dictionary where the keys are a parent table object or a child table object and the
                    values are dictionaries holding key value pairs of the elements within the objects and the values
                    being the resulting column names.
                    e.g. {'parent_object': {'parent_object.element': 'element_column_name'}}
                root_node:
                    Optional[str] : name of any root nodes of the data e.g. if data is:
                    {"root_el": {"orders": {"order": [{}]}}}, then root_node should be root_el.orders.order
                default_parse_list:
                    Optional[ParseListStrategy] : The default strategy of the parser to parse lists
                default_parse_dict:
                    Optional[ParseDictStrategy] : The default strategy of the parser to parse dictionaries
                parse_list_mapping:
                    Optional[Dict[str,ParseListStrategy]] :  A dictionary holding key value pairs, where the keys are
                    specific elements that are Lists in the JSON data and values being their Parsing strategy.
                    This is used when a specific list should be parsed differently than the default strategy.
                parse_dict_mapping:
                    Optional[Dict[str,ParseDictStrategy]] :  A dictionary holding key value pairs, where the keys are
                    specific elements that are Dictionaries in the JSON data and values being their Parsing strategy.
                    This is used when a specific dictionary should be parsed differently than the default strategy.
                ignore_undefined_tables:
                    Optional[bool] : If True, tables that are not defined in child or parent table definitions will not
                     be processed or output by the parser.
                ignore_undefined_columns:
                    Optional[bool] : If True, columns that are not defined in table column mapping will not be
                    processed or output by the parser.
                always_array:
                    Optional[List] : List of element names that should always be treated as arrays/lists even though
                    they might sometimes be of a different.
            Raises:
                JSONParserError - on parsing errors.
            """

    def __init__(self,
                 json_configuration_file: str = None,
                 configuration_dict: Dict = None,
                 parent_table: Dict[str, str] = None,
                 child_tables: Dict[str, str] = None,
                 table_primary_keys: Dict[str, List[str]] = None,
                 table_column_mapping: Dict[str, Dict[str, str]] = None,
                 root_node: str = "",
                 default_parse_list: ParseListStrategy = None,
                 default_parse_dict: ParseDictStrategy = None,
                 parse_list_mapping: Dict[str, ParseListStrategy] = None,
                 parse_dict_mapping: Dict[str, ParseDictStrategy] = None,
                 ignore_undefined_tables: bool = False,
                 ignore_undefined_columns: bool = False,
                 always_array: List = None) -> None:

        self.parent_table = parent_table
        self.child_tables = child_tables
        self.table_primary_keys = table_primary_keys
        self.root_node = root_node
        self.table_column_mapping = table_column_mapping
        self.default_parse_list = default_parse_list
        self.default_parse_dict = default_parse_dict
        self.parse_list_mapping = parse_list_mapping
        self.parse_dict_mapping = parse_dict_mapping
        self.ignore_undefined_tables = ignore_undefined_tables
        self.ignore_undefined_columns = ignore_undefined_columns
        self.always_array = always_array

        if self.parent_table:
            self.all_tables = self.parent_table.copy()
        if self.child_tables and self.parent_table:
            self.all_tables.update(self.child_tables)

        self._set_variables(json_configuration_file, configuration_dict)

    def _set_variables(self, json_file, configuration_dict):
        if json_file:
            self._set_variables_from_json_file(json_file)
        elif configuration_dict:
            self._set_variables_from_dict(configuration_dict)

    def _set_variables_from_json_file(self, json_file_path):
        with open(json_file_path) as json_file:
            config = json.load(json_file)
        self._set_variables_from_dict(config)

    def _set_variables_from_dict(self, configuration_dict):
        self.parent_table = configuration_dict.get("parent_table")
        if not self.parent_table:
            self.parent_table = {}
        self.child_tables = configuration_dict.get("child_tables")
        if not self.child_tables:
            self.child_tables = {}
        self.table_primary_keys = configuration_dict.get("table_primary_keys")
        if not self.table_primary_keys:
            self.table_primary_keys = {}
        self.root_node = configuration_dict.get("root_node")

        self.table_column_mapping = configuration_dict.get("table_column_mapping")

        self.all_tables = self.parent_table.copy()
        if self.child_tables:
            self.all_tables.update(self.child_tables)

        default_parse_list = configuration_dict.get("default_parse_list")
        if default_parse_list:
            default_parse_list = convert_str_to_parse_list_option(default_parse_list)
        default_parse_dict = configuration_dict.get("default_parse_dict")
        if default_parse_dict:
            default_parse_dict = convert_str_to_parse_dict_option(default_parse_dict)
        parse_list_mapping = configuration_dict.get("parse_list_mapping")
        if parse_list_mapping:
            parse_list_mapping = self.map_list_parsing_options(parse_list_mapping)
        parse_dict_mapping = configuration_dict.get("parse_dict_mapping")
        if parse_dict_mapping:
            parse_dict_mapping = self.map_dict_parsing_options(parse_dict_mapping)

        self.default_parse_list = default_parse_list or DEFAULT_PARSE_LIST
        self.default_parse_dict = default_parse_dict or DEFAULT_PARSE_DICT
        self.parse_list_mapping = parse_list_mapping
        self.parse_dict_mapping = parse_dict_mapping

        self.ignore_undefined_tables = configuration_dict.get("ignore_undefined_tables")
        self.ignore_undefined_columns = configuration_dict.get("ignore_undefined_columns")

        self.always_array = configuration_dict.get("always_array")

    @staticmethod
    def map_list_parsing_options(parse_list_mapping):
        for key in parse_list_mapping:
            parse_list_mapping[key] = convert_str_to_parse_list_option(parse_list_mapping[key])
        return parse_list_mapping

    @staticmethod
    def map_dict_parsing_options(parse_dict_mapping):
        for key in parse_dict_mapping:
            parse_dict_mapping[key] = convert_str_to_parse_dict_option(parse_dict_mapping[key])
        return parse_dict_mapping



class JSONParserException(Exception):
    pass


class InvalidRootNodeException(JSONParserException):
    DEFAULT_MESSAGE = "Invalid root node, could not parse JSON file"

    def __init__(self, msg=DEFAULT_MESSAGE, *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class MissingDataException(JSONParserException):
    DEFAULT_MESSAGE = "No data was found based on the specified root node"

    def __init__(self, msg=DEFAULT_MESSAGE, *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class InconsistentDataTypesException(JSONParserException):
    def __init__(self, msg=None, column=None, parse_strategy=None, processed_columns=None, *args, **kwargs):
        if not msg and column and parse_strategy and processed_columns:
            msg = f"""Object {column.object_name} has an inconsistent data parsing strategy once the type/strategy was '{parse_strategy.value}' another time type/strategy was '{processed_columns[column.object_name].value}'. If there are inconsistant data types use the 'always_array' option in the configuration of the JSONParser"""
        super().__init__(msg, *args, **kwargs)



from enum import Enum, auto
from typing import List, Dict
from .configuration import ParseListStrategy, ParseDictStrategy, ParserConfiguration
from .value_object import Column


class ObjectParseStrategy(Enum):
    CHILD_TABLE = "list of dictionaries"
    DICT_TO_COLUMNS = "dictionary (to columns)"
    DICT_TO_STRING = "dictionary (to string)"
    UNDEFINED_CHILD_TABLE_TO_IGNORE = "ignored table"
    LIST_TO_COLUMNS = "list (to columns)"
    LIST_TO_ROWS = "list (to rows)"
    LIST_TO_STRING = "list (to string)"
    TO_STRING_COLUMN = "simple datatype"


def get_parse_strategy(column: Column, parser_config: ParserConfiguration) -> ObjectParseStrategy:
    object_is_dict = _object_is_dict(column.data)
    object_is_list = _object_is_list(column.data)

    always_array = parser_config.always_array
    column_is_array = bool(always_array and column.object_name in always_array)

    mapping_direction = _get_mapping_direction(parser_config, column)

    parse_strategy = ObjectParseStrategy.TO_STRING_COLUMN

    if _is_object_child_table(column, parser_config.all_tables):
        parse_strategy = ObjectParseStrategy.CHILD_TABLE
    elif object_is_dict and column_is_array and not parser_config.ignore_undefined_tables:
        parse_strategy = ObjectParseStrategy.CHILD_TABLE
    elif _is_list_of_dicts(column.data) and parser_config.ignore_undefined_tables:
        parse_strategy = ObjectParseStrategy.UNDEFINED_CHILD_TABLE_TO_IGNORE
    elif _is_list_of_dicts(column.data) and not parser_config.ignore_undefined_tables:
        parse_strategy = ObjectParseStrategy.CHILD_TABLE
    elif object_is_dict and mapping_direction == ParseDictStrategy.TO_COLUMNS:
        parse_strategy = ObjectParseStrategy.DICT_TO_COLUMNS
    elif object_is_dict and mapping_direction == ParseDictStrategy.TO_STR:
        parse_strategy = ObjectParseStrategy.DICT_TO_STRING
    elif (object_is_list or column_is_array) and mapping_direction == ParseListStrategy.TO_COLUMNS:
        parse_strategy = ObjectParseStrategy.LIST_TO_COLUMNS
    elif (object_is_list or column_is_array) and mapping_direction == ParseListStrategy.TO_ROWS:
        parse_strategy = ObjectParseStrategy.LIST_TO_ROWS
    elif (object_is_list or column_is_array) and mapping_direction == ParseListStrategy.TO_STR:
        parse_strategy = ObjectParseStrategy.LIST_TO_STRING
    return parse_strategy


def _get_mapping_direction(parser_config, column):
    parse_list_mapping = parser_config.parse_list_mapping
    mapping_direction = ""
    parse_dict_mapping = parser_config.parse_dict_mapping
    if parse_dict_mapping and column.object_name in list(parse_dict_mapping.keys()):
        mapping_direction = parse_dict_mapping[column.object_name]
    elif parse_list_mapping and column.object_name in list(parse_list_mapping.keys()):
        mapping_direction = parse_list_mapping[column.object_name]
    elif parse_list_mapping and column.object_name in list(parse_list_mapping.keys()):
        mapping_direction = parse_list_mapping[column.object_name]
    elif _object_is_dict(column.data):
        mapping_direction = parser_config.default_parse_dict
    elif _object_is_list(column.data):
        mapping_direction = parser_config.default_parse_list

    return mapping_direction


def _object_is_dict(_object) -> bool:
    return isinstance(_object, Dict)


def _object_is_list(_object) -> bool:
    return isinstance(_object, List)


def _is_object_child_table(column: Column, all_tables) -> bool:
    return column.object_name in list(all_tables.keys())


def _is_list_of_dicts(object_) -> bool:
    return all(isinstance(i, dict) for i in object_) if isinstance(object_, List) else False



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="keboola.json-to-csv",
    version="0.0.2",
    author="Keboola KDS Team",
    setup_requires=['flake8'],
    tests_require=[],
    install_requires=[],
    author_email="support@keboola.com",
    description="General utility library for Python applications running in Keboola Connection environment",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/keboola/python-utils",
    packages=['keboola.json_to_csv', 'keboola.json_to_csv.configuration', 'keboola.json_to_csv.exception',
              'keboola.json_to_csv.value_object'],
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    test_suite='tests',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Education",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    python_requires='>=3.7'
)



from itertools import chain
from typing import Dict

from .configuration import ParserConfiguration


def normalize_parsed_data(configuration: ParserConfiguration, data: Dict) -> Dict:
    data = _fill_in_missing_keys_to_data(data)
    if configuration.table_column_mapping:
        data = _parse_data_with_specified_fields(configuration, data)
    else:
        data = _remove_parents_from_data(data)
    data = _normalize_headers(data)
    data = _fix_table_names(configuration, data)
    return data


def _remove_parents_from_data(csv_data):
    for data_name in csv_data:
        for i, row in enumerate(csv_data[data_name]):
            for key in list(csv_data[data_name][i].keys()):
                new_key = key.replace(f"{data_name}.", "")
                csv_data[data_name][i][new_key] = csv_data[data_name][i].pop(key)
    return csv_data


def _parse_data_with_specified_fields(configuration, csv_data):
    for data_name in csv_data:
        if current_table_fields := configuration.table_column_mapping.get(data_name):
            # TODO raise exception if current_table_fields does not exist
            for i, row in enumerate(csv_data[data_name]):
                for key in list(csv_data[data_name][i].keys()):
                    if key in current_table_fields:
                        csv_data[data_name][i][current_table_fields[key]] = csv_data[data_name][i].pop(key)
                    elif not configuration.ignore_undefined_columns:
                        new_key = key.replace(f"{data_name}.", "")
                        csv_data[data_name][i][new_key] = csv_data[data_name][i].pop(key)
                    else:
                        csv_data[data_name][i].pop(key)
    return csv_data


def _normalize_headers(csv_data):
    for data_name in csv_data:
        for i, row in enumerate(csv_data[data_name]):
            for key in list(row.keys()):
                csv_data[data_name][i][key.replace(".", "_")] = csv_data[data_name][i].pop(key)
    return csv_data


def _fix_table_names(configuration, csv_data):
    for data_name in list(csv_data.keys()):
        table_name = configuration.all_tables.get(data_name)
        if not table_name and configuration.ignore_undefined_tables:
            csv_data.pop(data_name)
            continue
        elif not table_name:
            table_name = f"{data_name.replace('.', '_')}.csv"
        csv_data[table_name] = csv_data.pop(data_name)
    return csv_data


def _fill_in_missing_keys_to_data(csv_data):
    for data_name in csv_data:
        keys = set(chain.from_iterable(csv_data[data_name]))
        for i, item in enumerate(csv_data[data_name]):
            csv_data[data_name][i].update({key: None for key in keys if key not in item})
    return csv_data



import logging
import json
import hashlib
from typing import List, Dict
from typing import Optional, Union
from .data_utils import normalize_parsed_data
from .configuration import ParserConfiguration
from .parse_strategy import ObjectParseStrategy, get_parse_strategy, ParseListStrategy, ParseDictStrategy
from .value_object import Column, Key
from .exception import InvalidRootNodeException, MissingDataException, InconsistentDataTypesException


class JSONParserError(Exception):
    pass


class JSONParser:
    def __init__(self,
                 json_configuration_file=None,
                 configuration_dict=None,
                 parent_table: Dict[str, str] = None,
                 table_primary_keys: Dict[str, List[str]] = None,
                 child_tables: Dict[str, str] = None,
                 table_column_mapping: Dict[str, Dict[str, str]] = None,
                 root_node: str = "",
                 default_parse_list: ParseListStrategy = None,
                 default_parse_dict: ParseDictStrategy = None,
                 parse_list_mapping=None,
                 parse_dict_mapping=None,
                 ignore_undefined_tables=False,
                 ignore_undefined_columns=False,
                 always_array=None) -> None:

        self.configuration = ParserConfiguration(json_configuration_file=json_configuration_file,
                                                 configuration_dict=configuration_dict,
                                                 parent_table=parent_table,
                                                 table_primary_keys=table_primary_keys,
                                                 child_tables=child_tables,
                                                 table_column_mapping=table_column_mapping,
                                                 root_node=root_node,
                                                 default_parse_list=default_parse_list,
                                                 default_parse_dict=default_parse_dict,
                                                 parse_list_mapping=parse_list_mapping,
                                                 parse_dict_mapping=parse_dict_mapping,
                                                 ignore_undefined_tables=ignore_undefined_tables,
                                                 ignore_undefined_columns=ignore_undefined_columns,
                                                 always_array=always_array)
        self.warnings = {}
        self.processed_columns = {}

    def parse_data(self, data: Dict) -> Dict:
        """

        Args:
            data (Dict): A dictionary containing the data to be parsed.

        Returns:
            parsed_data (Dict) : dictionary of parsed data with key value pairs, where keys are names of csv files, and
            values are lists of flat dictionaries.

        """
        self._reset_warnings()
        self._reset_processed_columns()
        data_to_parse = self._get_data_to_parse(data, self.configuration.root_node)
        parsed_data = {}
        for row in data_to_parse:
            parsed_row = self._parse_row_to_tables(row)
            for table_name in parsed_row:
                if table_name not in parsed_data:
                    parsed_data[table_name] = []
                parsed_data[table_name].extend(parsed_row[table_name])
        self._log_warnings()
        parsed_data = normalize_parsed_data(self.configuration, parsed_data)
        return parsed_data

    def _get_primary_keys(self, data: Dict, parent_table: str) -> List[Key]:
        primary_key_definitions = self.configuration.table_primary_keys.get(parent_table)
        if primary_key_definitions:
            primary_keys = self._get_primary_key_values(primary_key_definitions, data, parent_table)
        else:
            self.warnings[f"{parent_table}.pkey"] = f"Object {parent_table} does not have primary keys set, " \
                                                    f"generating them automatically"
            primary_keys = [self._generate_primary_key_from_data(parent_table, data)]
        return primary_keys

    @staticmethod
    def _get_primary_key_values(primary_key_definitions, data, parent_table):
        primary_keys = []
        key_prefix = "".join([parent_table, "."])
        for primary_key_definition in primary_key_definitions:
            if primary_key_definition.replace(key_prefix, "") in data:
                key_data = data[primary_key_definition.replace(key_prefix, "")]
                primary_key = Key(name=primary_key_definition, value=key_data, table=parent_table)
                primary_keys.append(primary_key)
        return primary_keys

    def _generate_primary_key_from_data(self, parent_table: str, data: Dict) -> Key:
        hash_key = "generated_pkey"
        if parent_table:
            hash_key = ".".join([parent_table, hash_key])
        hash_value = f"key_{self._hash_data_row(data)}"
        return Key(name=hash_key, value=hash_value, table=parent_table)

    @staticmethod
    def _hash_data_row(data_row: Dict) -> str:
        dict_str = json.dumps(data_row, sort_keys=True, ensure_ascii=True, default=str)
        return hashlib.sha256(dict_str.encode("utf-8")).hexdigest()

    def _parse_row_to_tables(self, data_object: Dict) -> Dict:
        row_data = {}
        parent_table = list(self.configuration.parent_table.keys())[0]
        row_data = self._parse_nested_dict(row_data, data_object, parent_table)
        return row_data

    def _parse_nested_dict(self, row_data: Dict, data: Dict, parent_table: str = "", table_index: int = 0,
                           foreign_keys: Optional[List[Key]] = None) -> Dict:

        if not foreign_keys:
            foreign_keys = []
        primary_keys = self._get_primary_keys(data, parent_table)

        if parent_table not in row_data:
            row_data[parent_table] = []

        for index, object_name in enumerate(data):
            if self._turn_dict_to_list(object_name, parent_table, data[object_name]):
                data[object_name] = [data[object_name]]
            column = Column(object_name, parent_table, data[object_name])

            parse_strategy = get_parse_strategy(column, self.configuration)
            self._check_object_consistency(column, parse_strategy)
            row_data = self._process_object(row_data, index, column, parse_strategy, primary_keys, foreign_keys,
                                            table_index)

        for i, primary_key in enumerate(primary_keys):
            if primary_key.name not in self.configuration.table_primary_keys.get(parent_table, []):
                primary_key_column = Column("generated_pkey", parent_table, primary_keys[i].value)
                row_data = self._parse_primitive(row_data, primary_key_column, 1, foreign_keys)

        return row_data

    def _check_object_consistency(self, column: Column, parse_strategy: ObjectParseStrategy) -> None:
        if column.object_name not in list(self.processed_columns.keys()):
            self.processed_columns[column.object_name] = parse_strategy
        elif self.processed_columns[column.object_name] != parse_strategy:
            raise InconsistentDataTypesException(column=column,
                                                 parse_strategy=parse_strategy,
                                                 processed_columns=self.processed_columns)

    def _process_object(self, row_data: Dict, index: int, column: Column, parse_strategy: ObjectParseStrategy,
                        primary_keys: List[Key], foreign_keys: List[Key], table_index: int) -> Dict:

        if parse_strategy == ObjectParseStrategy.CHILD_TABLE:
            row_data = self._parse_list_of_dicts(row_data, column, primary_keys, foreign_keys)
        elif parse_strategy == ObjectParseStrategy.DICT_TO_COLUMNS:
            row_data = self._flatten_simple_dict(row_data, column, table_index, foreign_keys)
        elif parse_strategy == ObjectParseStrategy.DICT_TO_STRING:
            column.data = json.dumps(column.data)
            row_data = self._parse_primitive(row_data, column, index, foreign_keys)
        elif parse_strategy == ObjectParseStrategy.UNDEFINED_CHILD_TABLE_TO_IGNORE:
            self.warnings[column] = f'Warning : Possible table "{column}" will be ignored as it is not specified ' \
                                    "in the configuration of the parser."
        elif parse_strategy == ObjectParseStrategy.LIST_TO_COLUMNS:
            column_list = [Column(f"{column.name}.{str(i)}", column.table, item) for i, item in
                           enumerate(column.data)]
            for column_row in column_list:
                row_data = self._parse_primitive(row_data, column_row, index, foreign_keys)
        elif parse_strategy == ObjectParseStrategy.LIST_TO_ROWS:
            if column.object_name not in row_data:
                row_data[column.object_name] = []
            for column_data in column.data:
                sub_column = Column(column.name, column.object_name, column_data)
                row_data = self._parse_primitive(row_data, sub_column, 0, primary_keys)
        else:
            row_data = self._parse_primitive(row_data, column, index, foreign_keys)
        return row_data

    def _parse_list_of_dicts(self, row_data: Dict, column: Column, primary_keys: List[Key],
                             foreign_keys: List[Key]) -> Dict:
        primary_keys.extend(foreign_keys)
        for index, d in enumerate(column.data):
            row_data = self._parse_nested_dict(row_data, d, column.object_name, foreign_keys=primary_keys,
                                               table_index=index)
        return row_data

    @staticmethod
    def _parse_primitive(row_data: Dict, column: Column, index: int, foreign_keys: List[Key]) -> Dict:

        if column.table not in row_data:
            row_data[column.table] = []

        if index == 0:
            row_data[column.table].append({})

        table_size = len(row_data[column.table])
        if foreign_keys:
            for foreign_key in foreign_keys:
                row_data[column.table][table_size - 1][foreign_key.name] = foreign_key.value
        row_data[column.table][table_size - 1][column.object_name] = column.data
        return row_data

    @staticmethod
    def _flatten_simple_dict(row_data: Dict, column: Column, index: int, foreign_keys: List[Key]) -> Dict:
        for d_key in column.data:
            new_key = f"{column.object_name}.{d_key}"
            if len(row_data[column.table]) < index + 1:
                row_data[column.table].append({})
            row_data[column.table][index][new_key] = column.data[d_key]

        if foreign_keys:
            for foreign_key in foreign_keys:
                row_data[column.table][index][foreign_key.name] = foreign_key.value
        return row_data

    def _column_is_always_array(self, parent_table: str, column: str) -> bool:
        column_definition = f"{parent_table}.{column}"
        if column_definition in list(self.configuration.all_tables.keys()):
            return True
        always_array = self.configuration.always_array
        if not always_array:
            return False
        return bool(always_array and column_definition in always_array)

    @staticmethod
    def _get_data_to_parse(data: Dict, root_node: str) -> List[Dict]:
        root_nodes = root_node.split(".")
        if len(root_nodes) == 1 and not root_nodes[0]:
            raise InvalidRootNodeException()
        try:
            for root_node in root_nodes:
                data = data.get(root_node)
        except AttributeError as attr_err:
            raise InvalidRootNodeException() from attr_err

        if not data:
            raise MissingDataException()

        if not isinstance(data, List):
            return [data]
        return data

    # TODO rename to something more understandable
    def _turn_dict_to_list(self, object_name: str, parent_table: str, object_data: Union[List, Dict]) -> bool:
        return bool(not isinstance(object_data, List) and self._column_is_always_array(parent_table, object_name))

    def _reset_warnings(self) -> None:
        self.warnings = {}

    def _reset_processed_columns(self) -> None:
        self.processed_columns = {}

    def _log_warnings(self) -> None:
        for warning in self.warnings:
            logging.warning(self.warnings[warning])
